# FINDE: Neural Differential Equations for Finding and Preserving Invariant Quantities

Neural networks have shown promise for modeling dynamical systems from data. Recent models, such as Hamiltonian neural networks, have been designed to ensure known geometric structures of target systems and have shown excellent modeling accuracy. However, in most situations where neural networks learn unknown systems, their underlying structures are also unknown. Even in such cases, one can expect that target systems are associated with first integrals (a.k.a. invariant quantities), which are quantities remaining unchanged over time. First integrals come from the conservation laws of system energy, momentum, and mass, from constraints on states, and from other features of governing equations. By leveraging projection methods and discrete gradient methods, we propose first integral-preserving neural differential equations (FINDE). The proposed FINDE finds and preserves first integrals from data, even in the absence of prior knowledge about the underlying structures. Experimental results demonstrate that the proposed FINDE is able to predict future states of given systems much longer and find various quantities consistent with well-known first integrals of the systems in a unified manner.

## Requirements

- Python v3.8.12
- scipy v1.7.3,
- pytorch v1.10.2
- torchdiffeq v0.1.1
- functorch v1.10 preview
- gplearn v0.4.2

The codes were made by modifying the codes for [Hamiltonian neural networks](https://github.com/greydanus/hamiltonian-nn) and [DGNet](https://github.com/tksmatsubara/discrete-autograd).

## Training and Evaluation

Commands to reproduce results:

```sh
python train_ode.py --name 2body          --model hnn    --norm --total_steps 100000 --finde continuous    --finde_num 2 --seed 0 --verbose
python train_pde.py --name KdV            --model cnn    --norm --total_steps 30000  --finde continuousPDE --finde_num 3 --seed 0 --verbose
python train_ode.py --name 2pend          --model sonode --norm --total_steps 100000 --finde continuous    --finde_num 5 --seed 0 --verbose
python train_ode.py --name FitzHughNagumo --model node   --norm --total_steps 30000  --finde continuous    --finde_num 2 --seed 0 --verbose
```

- `--name`: dataset. 2body, 2pend, FitzHughNagumo, KdV.
- `--model`: base model. hnn, node, sonode.
- `--seed`: random see.
- `--total_steps`: number of iterations.
- `--adjoint`: use adjoint method for PDEs.
- `--norm`: data normalization at the first layer.
- `--finde`: use FINDE.
- `--finde_num`: number of first integrals.
